/** @file
 ** Script program starter, shbang interpreter.
 *
 * This program takes a script filename as argument, possibly followed
 * by the script arguments.
 * The first line of the script file should be in the form of a Unix
 * shbang #! (sharp bang).
 * If StartScript find these characters, it take the remainder of the line
 * as the path of a script interpreter executable, plus additional parameters
 * for this interpreter.
 * It will concatenate the name of the script and its arguments to this path,
 * and run this command line.
 *
 * In other words, it will act like a Unix shell, running a script interpreter
 * indicated by the script file itself.
 * Note that you can omit the path, in the shbang line, as long as the program
 * is in the shell path.
 * StartScript verifies the existence of the exe; if not found (or if no shbang
 * line), it will fall on a default call.
 * You must enclose the path in double quotes if it has spaces.
 *
 * I made StartScript because I am starting to have several specialized Lua
 * interpreters (one for Win32, one for testing an API, plus the official
 * Lua 4.0, and the experimental Lua 5.0, etc.).
 * The classical (Windows) way to handle these various executables would be
 * to create a separate extension for each (eg. wlua, foolua, lua4, lua5...)
 * and make an assocation (in registry) for each of them.
 * This pollutes the namespace of HKEY_CLASSES_ROOT and brings the risk of
 * extension collision with another program (installed or future, Lua related
 * or not).
 * So I chose to use the shbang, and since I already have a lot of scripts
 * without shbang lines, I added a mechanism to retreive a default script
 * if none is given, or if the given one is not good (not found in path).
 * And while I was at it, I made it versatile enough to support any number
 * of unrelated interpreters (whatever scripting language).
 *
 * Suggested usage:
 * Previously, I associated in the registry (HKEY_CLASSES_ROOT)
 * the .lua extension to luafile which had the form:
 *
@verbatim
[HKEY_CLASSES_ROOT\luafile]
@="Lua: program"
"AlwaysShowExt"=""

[HKEY_CLASSES_ROOT\luafile\DefaultIcon]
@="D:\\ProgCmdLine\\Lua.exe,-1"

[HKEY_CLASSES_ROOT\luafile\shell\open\command]
@="\"D:\\ProgCmdLine\\Lua.exe\" -c -f \"%1\" %*"
@endverbatim
 *
 * With StartScript, I changed the command to:
 *
@verbatim
[HKEY_CLASSES_ROOT\luafile\shell\open\command]
@="\"D:\\ProgCmdLine\\ScriptStart.exe\" \"%1\" %*"
@endverbatim
 *
 * With this registry key, you can start scripts from the Explorer
 * by double-clicking on them, hitting Return when selected, dragging
 * file on them, etc.
 *
 * On WinNT/2k/XP, with cmd.exe command line interpreter, you can also
 * start scripts from the command line, just by typing the name of
 * the script and its parameters: FooBar.lua infile.txt outfile.txt
 *
 * On Dos-based Win9x/Me with command.com, you are out of luck since
 * Dos will not look in the registry. As a workaround, you can rename
 * StartScript.exe to something shorter, like ss.exe or go.exe or even
 * g.exe. You can then start a script by typing something like:
 * go FooBar.lua infile.txt outfile.txt
 *
 * To use the default interpreter mechanism, put in the same directory
 * as the executable a text file named as the executable, but with a .cfg
 * extension.
 * For example, if you use StartScript.exe, you should have a StartScript.cfg
 * file defining the default interpreters, depending on the extension.
 * The format of this file is simple: each line as an extension (without dot)
 * followed by a tab and the path and command line arguments of the interpreter.
 **/
/* Project history:
 *  1.00.000  2002/10/04 (PL)  First release state.
 *  0.00.000  2002/07/22 (PL)  Creation
 */
/* Copyright notice: See the PhiLhoSoft Licence.txt file for details.
StartScript is distributed under the zlib/libpng license.
Copyright (c) 2002 Philippe Lhoste / PhiLhoSoft
*/
#include <stdio.h>
#include <stdlib.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

/* Max. shbang line size: a path and some parameters */
#define MAX_SHBANG_LINE_SIZE	(MAX_PATH + 20)
#define ERR_BAD_USE			255
#define ERR_CANNOT_RUN		2

/* Extension of the configuration file, to be appended to the application name */
#define CONFIG_EXT	"cfg"
/* Maximum length of script file extensions. Should take care of most cases :-) */
#define MAX_EXTENSION_LEN	8
#define ERROR_LENGTH	127

/*#define DEBUG_MODE*/
#ifdef DEBUG_MODE
#include <stdarg.h>

void DebugPrintf(char *format, ...)
{
	char buffer[2000];
	va_list pArguments;

	va_start(pArguments, format);
	vsprintf(buffer, format, pArguments);
	va_end(pArguments);
	OutputDebugString(buffer);
}
#endif

/**
 * Copy the source into the target at a given offset, until the end
 * of the string is meet, or until the string in the target reaches
 * maxLen number of characters.
 * The target is always '\0' terminated.
 * @return the length of the string in sTarget after the concatenation,
 * or -1 if there was an overflow. In this case, the length
 * of the string in sTarget is maxLen.
 *
 * It is similar to strncat, but faster because we give the position where to
 * start to copy the source string: no need to compute the end of the current
 * target. This feature is useful when multiple copies are needed, eg.:
 * tLen = stradd(t, s1, 0, tSize-1);
 * if (tLen >= 0)	// Optional, the function is safe if called with negative length (no op)
 * {
 * 	tLen = stradd(t, s2, tLen, tSize-1);
 * }
 * and so on.
 */
int stradd(
	char *sTarget,	/**< Where to write the result */
	char *source,	/**< The string to concatenate */
	int currentLen,	/**< Current length of the string in sTarget,
					 * ie. the offset where to start to copy data. */
	int maxLen)		/**< Max. length of the string to put in sTarget
					 * = size of sTarget buffer - 1 */
{
	int targetLen = currentLen;
	char *pSource = source;

	if (sTarget == NULL)
	{
		targetLen = 0;
		goto end;
	}
	if (source == NULL || *source == '\0' || currentLen < 0 || currentLen >= maxLen)
	{
		goto end;
	}
	while (targetLen < maxLen && *pSource != '\0')
	{
		sTarget[targetLen++] = *pSource++;
	}
	sTarget[targetLen] = '\0';
	if (*pSource != '\0')
	{
		/* Overflow */
		targetLen = -1;
	}

end:
	return targetLen;
}

/**
 * Extract from the given @a line the first found parameter,
 * unquoted if needed. The resulting string is put in the @a param buffer,
 * of size @a maxParamLen + 1.
 * @return the offset in the @a line, after the extracted parameter.
 */
int GetParam(
	char *line,			/**< The line containing the parameter to extract */
	char *param,		/**< Where to write the result: the first param.
						 * of the given line, unquoted */
	int maxParamLen)	/**< Max. length of the @a param buffer */
{
	BOOL bIsQuoted = FALSE;
	int lenParam = 0, lineOffset = 0;

	/* Skip whitespace */
	while (lenParam < maxParamLen &&
		isspace(line[lineOffset]))
	{
		lineOffset++;
	}
	/* Quoted string? */
	if (line[lineOffset] == '"')
	{
		lineOffset++;
		bIsQuoted = TRUE;
	}
	/* Loop while:
	 * - not at end of target buffer nor of source
	 * - not found ending quote if quoted string
	 * - not found space if not quoted
	 */
	while (lenParam < maxParamLen &&
		line[lineOffset] != '\0' &&
		!(bIsQuoted && line[lineOffset] == '"') &&
		!(!bIsQuoted && isspace(line[lineOffset])))
	{
		param[lenParam++] = line[lineOffset++];
	}
	param[lenParam] = '\0';
	if (line[lineOffset] == '"')
	{
		lineOffset++;
	}

	return lineOffset;
}

/**
 * Wrapper of fgets, removing the trailing EOL.
 * @return The length of the read line, or 0 if EOF or oher problem.
 */
int GetLine(
	FILE *file,		/**< File handle */
	char *buffer,	/**< Buffer to receive the read line */
	int bufSize)	/**< Maximum size of the @a buffer */
{
	char *pRes;
	int len = 0;

	buffer[0] = '\0';
	pRes = fgets(buffer, bufSize, file);
	if (pRes != NULL)
	{
		len = strlen(buffer);
		if (len >= 1 && buffer[len - 1] == '\n')
		{
			buffer[len - 1] = '\0';	/* Remove '\n' */
		}
	}
	return len;
}

/**
 * Search in the configuration file the default interpreter able to
 * execute the given script, using the extension as identifier.
 * @return TRUE if found.
 */
BOOL GetInterpreter(
	char *extension,	/**< Extension of the script file */
	char *interpreter,	/**< Where to write the result: the found
						 * path of the interpreter */
	int maxInterpreterLen)	/**< Max. length of the @a interpreter buffer */
{
	FILE *file;
	char sFilename[MAX_PATH + 1];
	char sLine[MAX_SHBANG_LINE_SIZE + MAX_EXTENSION_LEN + 1];
	char sExt[MAX_EXTENSION_LEN + 1];
	int lenLine, lenExt;
	BOOL bFound = FALSE;

	interpreter[0] = '\0';
	lenLine = GetModuleFileName(NULL, sFilename, MAX_PATH);
	strcpy(sFilename + lenLine - 3, CONFIG_EXT);	/* Replace .exe by .cfg */
	/* Open configuration file */
	file = fopen(sFilename, "rt");
#ifdef DEBUG_MODE
	DebugPrintf("File=%s, Handle=%08X\n", sFilename, file);
#endif
	if (file != NULL)
	{
		do
		{
			lenLine = GetLine(file, sLine, sizeof(sLine));
			if (lenLine > 0)
			{
				lenExt = GetParam(sLine, sExt, MAX_EXTENSION_LEN);
				if (_stricmp(sExt, extension) == 0)
				{
					bFound = TRUE;
					/* Skip all whitespace, in case user has put several... */
					while (isspace(sLine[lenExt]) && lenExt < lenLine)
					{
						lenExt++;
					}
					strncpy(interpreter, sLine + lenExt, maxInterpreterLen);
					interpreter[maxInterpreterLen] = '\0';
				}
			}
		} while (lenLine > 0 && !bFound);
		fclose(file);
	}
	else
	{
		/* Bad argument, file not found */
#ifdef DEBUG_MODE
		DebugPrintf("fopen error=(%d) %s\n", errno, strerror(errno));
#endif
	}
#ifdef DEBUG_MODE
	DebugPrintf("Interpreter=%s\n", interpreter);
#endif

	return bFound;
}

/**
 * Transform a GetLastError call to a readable message.
 */
void GetErrorMessage(
	char *sErrorMsg,	/**< Buffer to receive the message */
	int len				/**< Max. length of string, ie. of buffer - 1 */
)
{
	LPVOID mMsgBuf;
	DWORD errorCode;

	*sErrorMsg = '\0';
	errorCode = GetLastError();
	if (errorCode != 0)
	{
		FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			errorCode,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
			(LPTSTR)&mMsgBuf,
			0,
			NULL
		);
		/** Process any inserts in pMsgBuf */
		/* ... */
		/* Copy the string */
		strncpy(sErrorMsg, (char *)mMsgBuf, len - 1);
		sErrorMsg[len] = '\0';
		/* Free the buffer */
		LocalFree(mMsgBuf);
	}
}

#ifdef _CONSOLE
int main(
	int argc,
	char *argv[])
#else
/**********************************************************************
 * Windows main procedure.
 * I use a WinMain/Windows program instead of a main/console program
 * so we don't get a Dos/cmd box flashing briefly...
 * @return Code for Windows
 **********************************************************************/
int WINAPI WinMain(
	HINSTANCE hInstance,	/**< Instance handle */
	HINSTANCE hPrevInst,	/**< Sibbling application handle */
	LPSTR  lpCmdLine,		/**< Command line used to run this application */
	int    nCmdShow)		/**< Default display mode of the window */
#endif
{
	int errorLevel = EXIT_FAILURE;
	char *cmdLine = NULL;
	char sAppName[MAX_PATH + 1];
	char sScriptName[MAX_PATH + 1];
	FILE *file;
	char sShbangLine[MAX_SHBANG_LINE_SIZE];
	char sInterpreter[MAX_PATH + 1];
	char *interpreter = NULL;
	char *mCommand = NULL;
	STARTUPINFO startupInfo;
	PROCESS_INFORMATION processInformation;
	DWORD res;
	int lenCmdLine, lenScriptPath = 0, lenCommand = 0, lenCmd = 0;

#ifdef _CONSOLE
	UNREFERENCED_PARAMETER(argc);
	UNREFERENCED_PARAMETER(argv);
#else
	UNREFERENCED_PARAMETER(hInstance);
	UNREFERENCED_PARAMETER(hPrevInst);
	UNREFERENCED_PARAMETER(lpCmdLine);
	UNREFERENCED_PARAMETER(nCmdShow);
#endif

	GetModuleFileName(NULL, sAppName, MAX_PATH);
	cmdLine = GetCommandLine();
	lenCmdLine = strlen(cmdLine);
#ifdef DEBUG_MODE
	DebugPrintf("Command line=%s\n", cmdLine);
#endif
#ifdef _CONSOLE
	/* In console mode, the command line starts with the path
	 * of this exe. We skip it.
	 */
	lenScriptPath = GetParam(cmdLine, sScriptName, MAX_PATH);
	cmdLine += lenScriptPath;
	lenCmdLine -= lenScriptPath;
#endif
	/* Get the name of the script to run */
	lenScriptPath = GetParam(cmdLine, sScriptName, MAX_PATH);
	if (*sScriptName == '\0')	/* No parameter */
	{
		errorLevel = ERR_BAD_USE;
		goto end;
	}

	/* Open this script */
	file = fopen(sScriptName, "rt");
#ifdef DEBUG_MODE
	DebugPrintf("File=%s, Handle=%08X\n", sScriptName, file);
#endif
	if (file != NULL)
	{
		/* Read the first line */
		GetLine(file, sShbangLine, sizeof(sShbangLine));
		fclose(file);
	}
	else
	{
		/* Bad argument, file not found */
#ifdef DEBUG_MODE
		DebugPrintf("fopen error=(%d) %s\n", errno, strerror(errno));
#endif
		goto end;
	}
#ifdef DEBUG_MODE
	DebugPrintf("sShbangLine=%s\n", sShbangLine);
#endif

	if (sShbangLine[0] == '#' && sShbangLine[1] == '!')
	{
		DWORD result;
		char sDir[MAX_PATH + 1];
		char *dummy;

		/* Shbang line found */
		interpreter = sShbangLine + 2;
		/* Get the path of the script interpreter (without parameters) */
		lenCommand = GetParam(interpreter, sInterpreter, MAX_PATH);
		/* Check if the interpreter can be found in the path */
		result = SearchPath(NULL, sInterpreter, NULL, MAX_PATH, sDir, &dummy);
#ifdef DEBUG_MODE
		DebugPrintf("sInterpreter=%s, found=%d\n", sInterpreter, result);
#endif
		if (result == 0)
		{
			/* Not found, we must try the default interpreter */
			interpreter = NULL;
		}
	}
	if (interpreter == NULL)
	{
		char sExt[_MAX_EXT];

		_splitpath(sScriptName, NULL, NULL, NULL, sExt);
#ifdef DEBUG_MODE
		DebugPrintf("sScriptName=%s, ext=%s\n", sScriptName, sExt);
#endif
		if (GetInterpreter(sExt + 1, sShbangLine, sizeof(sShbangLine) - 1))
		{
			DWORD result;
			char sDir[MAX_PATH + 1];
			char *dummy;

			interpreter = sShbangLine;
			/* Get the path of the script interpreter (without parameters) */
			lenCommand = GetParam(interpreter, sInterpreter, MAX_PATH);
			/* Check if the interpreter can be found in the path */
			result = SearchPath(NULL, sInterpreter, NULL, MAX_PATH, sDir, &dummy);
#ifdef DEBUG_MODE
			DebugPrintf("sInterpreter=%s, found=%d\n", sInterpreter, result);
#endif
			if (result == 0)
			{
				/* Not found, we must try the default interpreter */
				interpreter = NULL;
			}
		}
	}
	if (interpreter != NULL)
	{
		/* The command to issue is the path of the interpreter
		 * and its parameters, plus a space, plus the path to
		 * the script file and its parameters, from the command line.
		 */
		lenCommand = strlen(interpreter) + 1 + lenCmdLine;
		mCommand = HeapAlloc(GetProcessHeap(), 0, lenCommand+1);
		/* Add to the shbang command and parameters the script name */
		lenCmd = stradd(mCommand, interpreter, 0, lenCommand);
		lenCmd = stradd(mCommand, " ", lenCmd, lenCommand);
		lenCmd = stradd(mCommand, sScriptName, lenCmd, lenCommand);
		/* Add to the parameters those given on the command line */
		lenCmd = stradd(mCommand, cmdLine + lenScriptPath, lenCmd, lenCommand);
	}
	else
	{
		errorLevel = ERR_CANNOT_RUN;
		fprintf(stderr, "[%s] No shbang line and/or [default] interpreter not found!\n", sAppName);
		goto end;
	}
#ifdef DEBUG_MODE
	DebugPrintf("CreateProcess(%s)\n", mCommand);
#endif
	/* Don't use ShellExecute, as it spawns another command box */
	ZeroMemory(&startupInfo, sizeof(startupInfo));
	startupInfo.cb = sizeof(startupInfo);
	ZeroMemory(&processInformation, sizeof(processInformation));
	if (!CreateProcess(NULL,
		mCommand,
		NULL,
		NULL,
		FALSE,
		0,
		NULL,
		NULL,
		&startupInfo,
		&processInformation))
	{
		char sError[128];

		errorLevel = EXIT_FAILURE;
		GetErrorMessage(sError, 127);
		fprintf(stderr, "[%s] Unable to execute the following command: %s -- Error is: %s\n", sAppName, mCommand, sError);
		goto end;
	}
	/* Wait until child process exits */
    res = WaitForSingleObject(processInformation.hProcess, INFINITE);
#ifdef DEBUG_MODE
	DebugPrintf("WaitForSingleObject=%d\n", res);
#endif
	if (res == WAIT_FAILED)
	{
		errorLevel = EXIT_FAILURE;
	}
	else
	{
		GetExitCodeProcess(processInformation.hProcess, &res);
		errorLevel = res;
	}

    /* Close thread and process handles */
    CloseHandle(processInformation.hThread);
    CloseHandle(processInformation.hProcess);

end:
	if (mCommand != NULL)
	{
		HeapFree(GetProcessHeap(), 0, mCommand);
	}
	if (errorLevel > 0)
	{
#ifdef DEBUG_MODE
		DebugPrintf("errorlevel=%d\n", errorLevel);
#endif
		ExitProcess(errorLevel);
	}
	return EXIT_SUCCESS;
}
